# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import hashlib
import logging
import os
import pathlib
import re
import signal
from typing import Callable, Dict, FrozenSet, List, Optional, Tuple

from aiohttp import hdrs, web
from aiohttp.helpers import BasicAuth
import bonsai
import cachetools

from . import errors


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
@web.middleware
async def middleware(
    request: web.Request, handler: Callable[[web.Request], web.Response]
) -> web.Response:
    backend = request.app[AuthBackend.KEY]
    await backend.check(request)
    return await handler(request)


# --------------------------------------------------------------------------------------
class AuthBackend:

    KEY = "dlrepo_auth_backend"
    AUTH_DISABLED = os.getenv("DLREPO_AUTH_DISABLED", "0") == "1"
    IGNORE_PASSWORDS = os.getenv("DLREPO_IGNORE_PASSWORDS", "0") == "1"
    LDAP_URL = os.getenv("DLREPO_LDAP_URL", "")
    LDAP_START_TLS = os.getenv("DLREPO_LDAP_START_TLS", "1") != "0"
    LDAP_TLS = LDAP_URL.startswith("ldaps://") or LDAP_START_TLS
    LDAP_USER_FILTER = os.getenv("DLREPO_LDAP_USER_FILTER", "")
    LDAP_BASE_DN = os.getenv("DLREPO_LDAP_BASE_DN", "")
    LDAP_GROUPS_FILTER = os.getenv("DLREPO_LDAP_GROUPS_FILTER", "")
    HTTP_LOGIN_HEADER = os.getenv("DLREPO_AUTH_HTTP_LOGIN_HEADER", "")
    HTTP_GROUPS_HEADER = os.getenv("DLREPO_AUTH_HTTP_GROUPS_HEADER", "")
    TIMEOUT = int(os.getenv("DLREPO_LDAP_TIMEOUT", "10"))
    MAX_CONNECTIONS = int(os.getenv("DLREPO_LDAP_MAX_CONNECTIONS", "4"))
    CACHE_SIZE = int(os.getenv("DLREPO_AUTH_CACHE_SIZE", "4096"))
    CACHE_TTL = int(os.getenv("DLREPO_AUTH_CACHE_TTL", "600"))

    def __init__(self):
        self.seed = os.urandom(32)
        self.semaphore = None  # created in init() to ensure it uses the running loop
        self.flush_caches()

    def flush_caches(self) -> None:
        self.group_acls = parse_group_acls()
        self.user_groups = parse_auth_file(self.seed)
        self.auth_cache = cachetools.TTLCache(self.CACHE_SIZE, self.CACHE_TTL)
        self.access_cache = cachetools.TTLCache(self.CACHE_SIZE, self.CACHE_TTL)

    async def init(self, app: web.Application) -> None:
        if self.AUTH_DISABLED:
            for line in BANNED_AUTH_DISABLED.strip().splitlines():
                LOG.critical("%s", line)
        elif self.IGNORE_PASSWORDS:
            for line in BANNED_IGNORED_PASSWORDS.strip().splitlines():
                LOG.critical("%s", line)
        asyncio.get_running_loop().add_signal_handler(signal.SIGHUP, self.flush_caches)
        self.semaphore = asyncio.Semaphore(self.MAX_CONNECTIONS)

    async def get_user_groups_from_ldap(self, login: str, password: str) -> List[str]:
        if not self.IGNORE_PASSWORDS:
            if not login or not password:
                return []

        client = bonsai.LDAPClient(self.LDAP_URL, tls=self.LDAP_TLS)

        # resolve login to LDAP DN
        filter_exp = self.LDAP_USER_FILTER.format(
            login=bonsai.escape_attribute_value(login)
        )
        async with client.connect(is_async=True, timeout=self.TIMEOUT) as conn:
            search_results = await conn.search(
                base=self.LDAP_BASE_DN,
                scope=bonsai.LDAPSearchScope.SUBTREE,
                filter_exp=filter_exp,
                attrlist=["dn"],
                timeout=self.TIMEOUT,
            )
            if len(search_results) == 0:
                LOG.debug("no such user: %s", login)
                return []
            if len(search_results) > 1:
                raise bonsai.LDAPError(f"multiple users with uid={login}")
            dn = str(search_results[0]["dn"])

        if not self.IGNORE_PASSWORDS:
            client.set_credentials("SIMPLE", user=dn, password=password)

        groups = []
        try:
            LOG.debug("checking credentials for %s on %s", login, self.LDAP_URL)
            filter_exp = self.LDAP_GROUPS_FILTER.format(
                login=bonsai.escape_attribute_value(login), dn=dn
            )
            async with client.connect(is_async=True, timeout=self.TIMEOUT) as conn:
                search_results = await conn.search(
                    base=self.LDAP_BASE_DN,
                    scope=bonsai.LDAPSearchScope.SUBTREE,
                    filter_exp=filter_exp,
                    attrlist=["cn"],
                    timeout=self.TIMEOUT,
                )
                for group in search_results:
                    groups.append(group["cn"][0])
        except bonsai.AuthenticationError as e:
            LOG.debug("authentication failed for %s: %s", login, e)
            return []

        return groups

    async def get_user_groups_from_file(self, login: str, password: str) -> List[str]:
        if self.user_groups:
            auth = login
            if not self.IGNORE_PASSWORDS:
                auth += f":{password}"
            auth_key = hashlib.sha256(self.seed + auth.encode("utf-8")).digest()
            groups = self.user_groups.get(auth_key, None)
            if groups is not None:
                return groups

        return []

    async def check_basic_auth(
        self, request: web.Request
    ) -> Tuple[Optional[str], FrozenSet[str]]:
        """
        Parse basic credentials from the HTTP Authorization header.

        :return:
            A tuple of (login, groups).
        """
        auth_header = request.headers.get(hdrs.AUTHORIZATION, "")
        if not auth_header:
            return None, frozenset()

        # prepend basic auth with random seed before hash to use as safe cache key
        auth_key = hashlib.sha256(self.seed + auth_header.encode("utf-8")).digest()
        if auth_key in self.auth_cache:
            return self.auth_cache[auth_key]

        try:
            auth = BasicAuth.decode(auth_header, encoding="utf-8")
            if not auth.login:
                return None, frozenset()
        except ValueError:
            return None, frozenset()

        login, password = auth.login, auth.password

        if self.LDAP_URL:
            # use a semaphore to limit the number of concurrent ldap connections
            async with self.semaphore:
                groups = await self.get_user_groups_from_ldap(login, password)
        else:
            groups = await self.get_user_groups_from_file(login, password)

        if not groups:
            LOG.debug("user %s does not exist or is not in any group", login)
            raise self.auth_error(request)

        groups = frozenset(groups)
        self.auth_cache[auth_key] = (login, groups)
        LOG.debug("user %s authenticated: groups %s", login, groups)
        return login, groups

    def auth_error(self, request: web.Request) -> web.HTTPError:
        if request.path.startswith("/v2/"):
            # docker registry errors must be JSON and must have specific HTTP headers
            return errors.Unauthorized()
        return errors.AuthenticationRequired()

    async def check(self, request: web.Request) -> None:
        if self.AUTH_DISABLED:
            return

        is_write = request.method not in ("GET", "HEAD")
        if not is_write:
            if request.path.startswith("/static/") or request.path == "/favicon.ico":
                # no need for authentication for these
                return

        if self.HTTP_LOGIN_HEADER and self.HTTP_GROUPS_HEADER:
            login = request.headers.get(self.HTTP_LOGIN_HEADER, None)
            groups = request.headers.get(self.HTTP_GROUPS_HEADER, "").strip()
            groups = frozenset(groups.split(","))

        else:
            login, groups = await self.check_basic_auth(request)

        # store the username for display in the access log
        request["dlrepo_user"] = login or "ANONYMOUS"
        if login is None:
            groups = frozenset({"ANONYMOUS"})

        if login is None and (is_write or request.path == "/v2/"):
            # anonymous write access is always denied
            # the docker registry API states that anonymous GET requests to /v2/
            # must receive a 401 unauthorized error
            # https://docs.docker.com/registry/spec/api/#api-version-check
            raise self.auth_error(request)

        # determine authenticated user acls from their groups
        acls = []
        for g in groups:
            for acl in self.group_acls.get(g, []):
                acls.append(acl.expand_user(login or ""))
        acls = frozenset(acls)

        # store user acls for reuse in views to filter out non-accessible elements
        request["dlrepo_user_acls"] = acls

        if not self.access_granted(
            acls, is_write, request.path, request["dlrepo_user"]
        ):
            raise self.auth_error(request)

    def access_granted(
        self,
        acls: FrozenSet["ACL"],
        is_write: bool,
        url: str,
        username: Optional[str] = None,
    ) -> bool:
        """
        Check if the provided ACLs give write and/or read access to the specified URL.
        """
        access_key = (acls, is_write, url)
        granted = self.access_cache.get(access_key)
        if granted is not None:
            return granted
        granted = False
        for acl in acls:
            if acl.access_granted(is_write, url):
                if username is not None:
                    LOG.debug("access granted to %s by ACL %s", username, acl)
                granted = True
                break
        self.access_cache[access_key] = granted
        return granted


# --------------------------------------------------------------------------------------
BANNED_AUTH_DISABLED = r"""
########################################
#                /!\                   #
# NO AUTHENTICATION! ANYONE CAN WRITE! #
########################################
"""
BANNED_IGNORED_PASSWORDS = r"""
#####################################
#                /!\                #
# PASSWORD AUTHENTICATION BYPASSED! #
#####################################
"""
AUTH_FILE_RE = re.compile(
    r"""
    ^
    (?P<user>[^:]+):
    (?P<password>.+):
    (?P<groups>[\w\s-]+)
    $
    """,
    re.MULTILINE | re.VERBOSE,
)


# --------------------------------------------------------------------------------------
def parse_auth_file(seed: bytes) -> Dict[bytes, List[str]]:
    user_groups = {}

    try:
        auth_file = os.getenv("DLREPO_AUTH_FILE", "/etc/dlrepo/auth")
        buf = pathlib.Path(auth_file).read_text(encoding="utf-8")
        for match in AUTH_FILE_RE.finditer(buf):
            groups = match.group("groups").strip().split()
            if not groups:
                continue
            auth = match.group("user")
            if not AuthBackend.IGNORE_PASSWORDS:
                auth += ":" + match.group("password")
            auth_key = hashlib.sha256(seed + auth.encode("utf-8")).digest()
            user_groups[auth_key] = groups

    except FileNotFoundError as e:
        if not AuthBackend.LDAP_URL and not AuthBackend.AUTH_DISABLED:
            LOG.error("DLREPO_AUTH_FILE: %s", e)

    return user_groups


# --------------------------------------------------------------------------------------
def parse_group_acls() -> Dict[str, FrozenSet["ACL"]]:
    group_acls = {}

    try:
        acls_dir = os.getenv("DLREPO_ACLS_DIR", "/etc/dlrepo/acls")
        for file in pathlib.Path(acls_dir).iterdir():
            if not file.is_file():
                continue
            acls = set()
            for line in file.read_text().splitlines():
                tokens = re.sub(r"#.*$", "", line).strip().split()
                if len(tokens) < 2:
                    continue
                access, *args = tokens
                if access not in ("ro", "rw"):
                    continue
                globs = []
                for a in args:
                    if a.startswith("!"):
                        globs.append((a[1:], True))
                    else:
                        globs.append((a, False))
                acls.add(ACL(access == "rw", frozenset(globs)))
            group_acls[file.name] = frozenset(acls)

    except FileNotFoundError as e:
        if not AuthBackend.AUTH_DISABLED:
            LOG.error("DLREPO_ACLS_DIR: %s", e)

    return group_acls


# --------------------------------------------------------------------------------------
class ACL:
    def __init__(self, read_write: bool, globs: FrozenSet[Tuple[str, bool]]):
        self.read_write = read_write
        self.globs = globs
        self.can_expand_user = any("$user" in g for g, _ in self.globs)
        self.patterns = [
            (self.translate_acl_glob(glob), invert) for glob, invert in self.globs
        ]

    def expand_user(self, login: str) -> "ACL":
        if not self.can_expand_user:
            return self
        globs = []
        for glob, invert in self.globs:
            globs.append((glob.replace("$user", login), invert))
        return ACL(self.read_write, frozenset(globs))

    def access_granted(self, is_write: bool, path: str) -> bool:
        if is_write and not self.read_write:
            return False
        granted = True
        for pattern, invert in self.patterns:
            if pattern.match(path):
                if invert:
                    granted = False
                    break
            else:
                if not invert:
                    granted = False
                    break
        return granted

    @staticmethod
    def translate_acl_glob(glob: str) -> re.Pattern:
        if glob and glob[0] == "~":
            return re.compile(glob[1:])
        pattern = r"^"
        i = 0
        while i < len(glob):
            c = glob[i]
            i += 1
            if c == "*":
                if i < len(glob) and glob[i] == "*":
                    i += 1
                    pattern += r".*"
                else:
                    pattern += r"[^/]*"
            elif c == "?":
                pattern += r"[^/]"
            else:
                pattern += re.escape(c)
        pattern += r"$"
        return re.compile(pattern)

    def __str__(self):
        access = "rw" if self.read_write else "ro"
        globs = []
        for glob, invert in self.globs:
            if invert:
                glob = f"!{glob}"
            globs.append(glob)
        return f"{access} {' '.join(globs)}"

    def __repr__(self):
        return f"ACL(read_write={self.read_write}, globs={self.globs})"

    def __eq__(self, other):
        if not isinstance(other, ACL):
            return False
        return other.read_write == self.read_write and other.globs == self.globs

    def __hash__(self):
        return hash((self.read_write, self.globs))



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio

from aiohttp import web
import aiohttp_jinja2

from .util import BaseView


# --------------------------------------------------------------------------------------
class TagView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}"
        yield "/branches/{branch}/{tag}/"
        yield "/~{user}/branches/{branch}/{tag}"
        yield "/~{user}/branches/{branch}/{tag}/"

    def _get_tag(self):
        match = self.request.match_info
        if match["tag"] in ("latest", "stable") and self.request.method not in (
            "GET",
            "HEAD",
        ):
            raise web.HTTPMethodNotAllowed(self.request.method, ["GET", "HEAD"])
        try:
            tag = self.repo().get_branch(match["branch"]).get_tag(match["tag"])
            if not tag.exists():
                raise web.HTTPNotFound()
            if tag.url() != self.request.path:
                raise web.HTTPFound(tag.url())
            return tag
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e

    async def get(self):
        """
        List tag contents.
        """
        tag = self._get_tag()
        data = {
            "tag": {
                "name": tag.name,
                "released": tag.is_released(),
                "locked": tag.is_locked(),
                "publish_status": tag.publish_status(),
                "jobs": [],
            },
        }
        for job in tag.get_jobs():
            if not self.access_granted(job.url()):
                continue
            job_data = job.get_metadata()
            if self.request.query:
                for key, value in self.request.query.items():
                    if key not in job_data or str(job_data[key]) != value:
                        break
                else:
                    data["tag"]["jobs"].append(job_data)
            else:
                data["tag"]["jobs"].append(job_data)
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template("tag.html", self.request, data)
        return web.json_response(data)

    async def put(self):
        """
        Change the released and/or locked status of a tag.
        """
        tag = self._get_tag()
        try:
            data = (await self.json_body())["tag"]
            released = data.get("released")
            if released is not None and not isinstance(released, bool):
                raise TypeError()
            locked = data.get("locked")
            if locked is not None and not isinstance(locked, bool):
                raise TypeError()
        except (TypeError, KeyError) as e:
            raise web.HTTPBadRequest(reason="invalid parameters") from e

        try:
            if released is not None:
                semaphore = self.request.app["dlrepo_publish_semaphore"]
                tag.set_released(released, semaphore)
            if locked is not None:
                tag.set_locked(locked)
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e

        return web.Response()

    async def delete(self):
        """
        Delete a tag and all its contents.
        """
        loop = asyncio.get_running_loop()
        try:
            tag = self._get_tag()
            force = "force" in self.request.query
            await loop.run_in_executor(None, lambda: tag.delete(force=force))
            self.repo().schedule_cleanup_orphans()
            return web.Response()
        except OSError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import json
from pathlib import Path
from typing import Dict, List, Tuple
import weakref

from .util import parse_digest


# --------------------------------------------------------------------------------------
class ContainerRegistry:
    def __init__(self, repo: "AbstractRepository"):
        self.repo = weakref.proxy(repo)

    def repositories(self) -> List[str]:
        repos = set()
        for d in self.repo.path().glob("branches/*/*/*/container"):
            if d.is_dir():
                job = d.parent.name
                branch = d.parent.parent.parent.name
                repos.add(f"{branch}/{job}")
        for d in self.repo.path().glob("products/*/*/*/*/container"):
            if d.is_dir():
                product_branch = d.parent.parent.name
                variant = d.parent.parent.parent.name
                product = d.parent.parent.parent.parent.name
                repos.add(f"{product}/{variant}/{product_branch}")
        return list(repos)

    def product_tags(self, product: str, variant: str, branch: str) -> List[str]:
        tags = set()
        glob = f"products/{product}/{variant}/{branch}/*/container"
        for d in self.repo.path().glob(glob):
            if d.is_dir():
                tags.add(d.parent.name)
        if tags:
            tags.add("latest")
        return sorted(tags)

    def job_tags(self, branch: str, job: str) -> List[str]:
        tags = set()
        glob = f"branches/{branch}/*/{job}/container"
        for d in self.repo.path().glob(glob):
            if d.is_dir():
                tags.add(d.parent.parent.name)
        if tags:
            tags.add("latest")
        return sorted(tags)

    def blob_link_path(self, parent_path: Path, digest: str):
        if not parent_path.is_dir():
            raise FileNotFoundError()
        algo, digest = parse_digest(digest.lower())
        return parent_path / f"container/blobs/{algo}/{digest[:2]}/{digest}"

    def manifest_by_digest(self, digest: str) -> Path:
        path = self.repo.blob_path(digest)
        if not path.is_file():
            raise FileNotFoundError()
        return path

    def manifest_by_parent(self, parent_path: Path) -> Tuple[Path, str]:
        if not parent_path.is_dir():
            raise FileNotFoundError()
        files = list(parent_path.glob("container/manifests/*/*/*"))
        if len(files) != 1:
            raise FileNotFoundError()
        if not files[0].is_file():
            raise FileNotFoundError()
        path = files[0]
        digest = path.name
        algo = path.parent.parent.name
        return (path, f"{algo}:{digest}")

    def _link_blob_to_job(self, digest: str, job: "Job") -> str:
        job_path = self.repo.blob_path(digest, parent=job.path() / "container/blobs")
        self.repo.link_blob(digest, job_path)
        return job_path.stat().st_size

    def new_manifest(self, job: "Job", manifest: Dict) -> str:
        # For an unknown reason, the manifest json file uploaded by "docker push"
        # does not specify a "size" field for every blob.
        # While "docker pull" can perfectly live with that and k8s/dockershim,
        # k8s/containerd on the other hand considers that the blob is of size 0
        # and issues an error like:
        #
        #   failed commit on ref "config-sha256:5b238[...]": commit failed:
        #   unexpected commit digest sha256:e3b0c[...], expected sha256:5b238[...]:
        #   failed precondition
        #
        # Indeed, e3b0c[...] is the sha256 of the empty string.
        #
        # To fix this, we add the missing 'size' fields into the manifest json
        # file right after it is uploaded. The blobs being uploaded before the
        # manifest, we can read their size on the filesystem.

        # link config blob into the job folder
        job.create()
        config = manifest.get("config", {})
        config["size"] = self._link_blob_to_job(config.get("digest"), job)

        # link layers blobs into the job folder
        for l in manifest.get("layers", []):
            l["size"] = self._link_blob_to_job(l.get("digest"), job)

        data = json.dumps(manifest).encode("utf-8")
        digest = "sha256:" + hashlib.sha256(data).hexdigest()

        manifest_path = self.repo.blob_path(digest)
        if not manifest_path.is_file():
            manifest_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
            manifest_path.write_bytes(data)

        job_manifest_path = self.repo.blob_path(
            digest, parent=job.path() / "container/manifests"
        )
        if job_manifest_path.is_file():
            job_manifest_path.unlink()
        self.repo.link_blob(digest, job_manifest_path)

        return digest



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import logging
import os
from pathlib import Path
import shutil
import time
from typing import Awaitable, Callable, Iterator, Optional

from cachetools import LFUCache

from .branch import Branch
from .container import ContainerRegistry
from .product import Product
from .tag import Tag
from .util import CHUNK_SIZE, file_digest, human_readable, parse_digest


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class AbstractRepository:
    def __init__(self, path: str):
        self._path = Path(path)
        self.parent = None
        self.container_registry = ContainerRegistry(self)

    def get_branches(self) -> Iterator[Branch]:
        yield from Branch.all(self)

    def get_branch(self, name: str) -> Branch:
        return Branch(self, name)

    def get_products(self) -> Iterator[Product]:
        yield from Product.all(self)

    def get_product(self, name: str) -> Product:
        return Product(self, name)

    def path(self) -> Path:
        return self._path

    def create(self):
        self._path.mkdir(mode=0o755, parents=True, exist_ok=True)

    def url_bit(self) -> Optional[str]:
        raise NotImplementedError()

    def schedule_cleanup_orphans(self):
        raise NotImplementedError()

    def schedule_cleanup_all(self):
        raise NotImplementedError()

    def next_upload(self) -> str:
        raise NotImplementedError()

    def blob_path(self, digest: str, parent: Optional[Path] = None) -> Path:
        raise NotImplementedError()

    def rmtree(self, path: Path):
        raise NotImplementedError()

    async def update_upload(
        self, uuid: str, stream: Callable[[int], Awaitable[bytes]]
    ) -> int:
        raise NotImplementedError()

    def cancel_upload(self, uuid: str):
        raise NotImplementedError()

    async def finalize_upload(self, uuid: str, digest: str) -> Path:
        raise NotImplementedError()

    def link_blob(self, digest: str, target: Path):
        raise NotImplementedError()

    def link_blob_ignore_quota(self, digest: str, target: Path):
        raise NotImplementedError()

    def copy_blob(self, digest: str, target: Path):
        raise NotImplementedError()


# --------------------------------------------------------------------------------------
class ArtifactRepository(AbstractRepository):
    def __init__(self, path: str):
        super().__init__(path)
        self.blobs = self._path / ".blobs"
        self.uploads = self._path / ".uploads"
        self.user_repos = LFUCache(maxsize=512)
        self.cleanup_task = None
        self.cleanup_timer = None
        self.need_cleanup = False
        self.cleanup_tags = False

    def get_user_repo(self, user: str) -> "UserRepository":
        user = user.lower()
        repo = self.user_repos.get(user)
        if repo is None:
            repo = self.user_repos[user] = UserRepository(self, user)
        return repo

    def get_user_repos(self) -> Iterator["UserRepository"]:
        yield from UserRepository.all(self)

    def url_bit(self) -> Optional[str]:
        return None

    def url(self) -> str:
        return "/"

    def _cleanup_timer_cb(self):
        self.schedule_cleanup_all()
        self.start_cleanup_timer()

    def start_cleanup_timer(self):
        self.cleanup_timer = asyncio.get_running_loop().call_later(
            self.CLEANUP_PERIOD, self._cleanup_timer_cb
        )

    def schedule_cleanup_orphans(self):
        self._schedule_cleanup(tags=False)

    def schedule_cleanup_all(self):
        self._schedule_cleanup(tags=True)

    def _schedule_cleanup(self, *, tags: bool):
        self.cleanup_tags = self.cleanup_tags or tags
        if self.cleanup_task:
            self.need_cleanup = True
        else:
            self.cleanup_task = asyncio.create_task(self._cleanup(self.cleanup_tags))
            self.cleanup_task.add_done_callback(self._cleanup_done_cb)
            self.need_cleanup = False
            self.cleanup_tags = False

    async def _cleanup(self, tags: bool):
        if tags:
            await self._cleanup_tags()
            self._refresh_user_quota()
        await self._cleanup_orphan_blobs()

    async def _cleanup_tags(self):
        loop = asyncio.get_running_loop()
        deleted = 0

        for branch in self.get_branches():
            released_tags = []
            daily_tags = []

            for tag in branch.get_tags():
                if tag.is_locked():
                    continue
                if tag.is_released():
                    released_tags.append(tag)
                else:
                    daily_tags.append(tag)

            released_tags.sort(key=Tag.creation_date, reverse=True)
            daily_tags.sort(key=Tag.creation_date, reverse=True)

            policy = branch.get_cleanup_policy()
            max_daily = policy.get("max_daily_tags", 0)
            if isinstance(max_daily, int) and max_daily > 0:
                for tag in daily_tags[max_daily:]:
                    LOG.info("Deleting daily tag %s/%s ...", branch.name, tag.name)
                    await loop.run_in_executor(None, tag.delete)
                    deleted += 1

            max_released = policy.get("max_released_tags", 0)
            if isinstance(max_released, int) and max_released > 0:
                for tag in released_tags[max_released:]:
                    LOG.info("Deleting released tag %s/%s ...", branch.name, tag.name)
                    await loop.run_in_executor(None, tag.delete, force=True)
                    deleted += 1

        LOG.info("Deleted %d tags", deleted)

    def _refresh_user_quota(self):
        LOG.debug("Refreshing all users quotas")
        for user_repo in self.get_user_repos():
            user_repo.disk_usage_refresh()
            user_repo.disk_usage_save()

    ORPHAN_BLOB_LIFETIME = int(os.getenv("DLREPO_ORPHAN_BLOB_LIFETIME", "600"))

    async def _cleanup_orphan_blobs(self):
        LOG.debug("Deleting orphan blobs")
        size = 0
        count = 0
        for folder in (self.blobs, self.uploads):
            if not folder.is_dir():
                return
            now = time.time()
            for root, _, files in os.walk(folder):
                for f in files:
                    await asyncio.sleep(0)
                    f = Path(root, f)
                    if not f.is_file():
                        continue
                    stat = f.stat()
                    if stat.st_nlink > 1:
                        continue
                    if now - stat.st_mtime < self.ORPHAN_BLOB_LIFETIME:
                        continue
                    f.unlink()
                    size += stat.st_size
                    count += 1
        LOG.info("Deleted %d orphan blobs (%s bytes)", count, human_readable(size))

    CLEANUP_PERIOD = int(os.getenv("DLREPO_TAG_CLEANUP_PERIOD", str(24 * 60 * 60)))

    def _cleanup_done_cb(self, task: asyncio.Task):
        self.cleanup_task = None
        exc = task.exception()
        if exc:
            LOG.error("exception in cleanup %r", task, exc_info=exc)
        if self.need_cleanup:
            self._schedule_cleanup(tags=False)

    def next_upload(self) -> str:
        self.uploads.mkdir(mode=0o755, parents=True, exist_ok=True)
        while True:
            try:
                path = self.uploads / os.urandom(16).hex()
                path.touch(mode=0o600, exist_ok=False)
                return path.name
            except FileExistsError:
                continue

    def upload_path(self, uuid: str):
        if "/" in uuid or ".." in uuid:
            raise FileNotFoundError(uuid)
        path = self.uploads / uuid
        if not path.is_file():
            raise FileNotFoundError(uuid)
        return path

    def blob_path(self, digest: str, parent: Optional[Path] = None) -> Path:
        if not digest:
            raise ValueError(f"invalid digest: {digest}")
        if parent is None:
            parent = self.blobs
        algo, digest = parse_digest(digest.lower())
        path = parent / f"{algo}/{digest[:2]}/{digest}"
        if path.is_file():
            path.touch()
        return path

    def rmtree(self, path: Path):
        shutil.rmtree(path, onerror=_rmtree_error_cb)

    async def update_upload(
        self, uuid: str, stream: Callable[[int], Awaitable[bytes]]
    ) -> int:
        return await _stream_to_file(stream, self.upload_path(uuid))

    def cancel_upload(self, uuid: str):
        self.upload_path(uuid).unlink()

    async def finalize_upload(self, uuid: str, digest: str) -> Path:
        return await _check_and_move(
            self.upload_path(uuid), self.blob_path(digest), digest
        )

    def link_blob(self, digest: str, target: Path):
        _copy_or_link(self.blob_path(digest), target)

    def link_blob_ignore_quota(self, digest: str, target: Path):
        _copy_or_link(self.blob_path(digest), target)

    def copy_blob(self, digest: str, target: Path):
        _copy_or_link(self.blob_path(digest), target, copy=True)


# --------------------------------------------------------------------------------------
class UserRepository(AbstractRepository):
    def __init__(self, base: ArtifactRepository, user: str):
        self.base = base
        self.user = user
        super().__init__(base.path() / "users" / user)
        self.disk_usage = self._parse_disk_usage_file()
        self.disk_usage_dirty = False

    @classmethod
    def all(cls, base: ArtifactRepository) -> Iterator["UserRepository"]:
        path = base.path() / "users"
        if not path.is_dir():
            return
        for d in path.iterdir():
            if not d.is_dir():
                continue
            yield cls(base, d.name)

    def _parse_disk_usage_file(self) -> int:
        du_path = self.path() / ".disk_usage"
        if du_path.is_file():
            try:
                return int(du_path.read_text(encoding="utf-8"))
            except ValueError:
                pass
        return 0

    def disk_usage_refresh(self):
        self.disk_usage = 0
        for root, _, files in os.walk(self.path()):
            for f in files:
                self.disk_usage += Path(root, f).stat().st_size
        self.disk_usage_dirty = True

    def disk_usage_save(self):
        if not self.disk_usage_dirty:
            return
        self.create()
        du_path = self.path() / ".disk_usage"
        du_path.write_text(str(self.disk_usage), encoding="utf-8")
        self.disk_usage_dirty = False

    # 10G max usage per user
    QUOTA = int(os.getenv("DLREPO_USER_QUOTA", str(10 * (1024**3))))

    def disk_usage_add(self, usage: int):
        new_usage = self.disk_usage + usage
        if new_usage > self.QUOTA:
            usage = f"{human_readable(self.disk_usage)}/{human_readable(self.QUOTA)}"
            raise PermissionError(
                f"User {self.user} quota exceeded ({usage}). Please make some ménage."
            )
        self.disk_usage = new_usage
        self.disk_usage_dirty = True

    def disk_usage_rm(self, usage: int):
        self.disk_usage = max(0, self.disk_usage - usage)
        self.disk_usage_dirty = True

    def url_bit(self) -> Optional[str]:
        return f"~{self.user}"

    def url(self) -> str:
        return f"/~{self.user}/"

    def schedule_cleanup_orphans(self):
        self.base.schedule_cleanup_orphans()

    def schedule_cleanup_all(self):
        self.base.schedule_cleanup_all()

    def next_upload(self) -> str:
        return self.base.next_upload()

    def cancel_upload(self, uuid: str):
        self.disk_usage_rm(self.base.upload_path(uuid).stat().st_size)
        try:
            self.base.cancel_upload(uuid)
        finally:
            self.disk_usage_save()

    def rmtree(self, path: Path):
        removed = 0
        for root, _, files in os.walk(path):
            for f in files:
                removed += Path(root, f).stat().st_size
        self.disk_usage_rm(removed)
        try:
            self.base.rmtree(path)
        finally:
            self.disk_usage_save()

    async def update_upload(
        self, uuid: str, stream: Callable[[int], Awaitable[bytes]]
    ) -> int:
        try:
            return await _stream_to_file(stream, self.base.upload_path(uuid), self)
        finally:
            self.disk_usage_save()

    async def finalize_upload(self, uuid: str, digest: str) -> Path:
        try:
            return await _check_and_move(
                self.base.upload_path(uuid), self.base.blob_path(digest), digest, self
            )
        finally:
            self.disk_usage_save()

    def blob_path(self, digest: str, parent: Optional[Path] = None) -> Path:
        return self.base.blob_path(digest, parent=parent)

    def link_blob(self, digest: str, target: Path):
        try:
            _copy_or_link(self.base.blob_path(digest), target, self)
        finally:
            self.disk_usage_save()

    def link_blob_ignore_quota(self, digest: str, target: Path):
        _copy_or_link(self.base.blob_path(digest), target)

    def copy_blob(self, digest: str, target: Path):
        _copy_or_link(self.base.blob_path(digest), target, copy=True)


# --------------------------------------------------------------------------------------
def _rmtree_error_cb(func, path, exc_info):
    # nothing much we can do here, simply log a message
    LOG.error("%s(%r) failed:", func, path, exc_info=exc_info)


async def _stream_to_file(
    stream: Callable[[int], Awaitable[bytes]],
    path: Path,
    user_repo: Optional[UserRepository] = None,
) -> int:
    # TODO: call blocking stuff (open, write, unlink) in a thread?
    try:
        with path.open("ab") as f:
            while True:
                chunk = await stream(CHUNK_SIZE)
                if not chunk:
                    break
                if user_repo is not None:
                    user_repo.disk_usage_add(len(chunk))
                f.write(chunk)
    except:
        if user_repo is not None:
            user_repo.disk_usage_rm(path.stat().st_size)
        path.unlink()
        raise
    return path.stat().st_size


# --------------------------------------------------------------------------------------
def _copy_or_link(
    src: Path, dst: Path, user_repo: Optional[UserRepository] = None, copy: bool = False
):
    if not src.is_file():
        raise FileNotFoundError()
    dst.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
    if dst.is_file():
        if user_repo is not None:
            user_repo.disk_usage_rm(dst.stat().st_size)
        dst.unlink()
    if user_repo is not None:
        user_repo.disk_usage_add(src.stat().st_size)
    if copy:
        shutil.copyfile(src, dst)
    else:
        os.link(src, dst)


# --------------------------------------------------------------------------------------
async def _check_and_move(
    src: Path, dst: Path, digest: str, user_repo: Optional[UserRepository] = None
):
    algo, dig = parse_digest(digest.lower())
    loop = asyncio.get_running_loop()
    if await loop.run_in_executor(None, file_digest, algo, src) != dig:
        if user_repo is not None:
            user_repo.disk_usage_rm(src.stat().st_size)
        src.unlink()
        try:
            os.removedirs(src.parent)
        except OSError:
            pass
        raise ValueError(f"Received data does not match digest: {digest}")
    dst.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
    src.rename(dst)
    dst.chmod(0o644)
    return dst



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

from http import HTTPStatus

from aiohttp import web

from . import errors
from .util import BaseView


# --------------------------------------------------------------------------------------
class ContainerView(BaseView):
    def response(self, body=None, path=None, status=HTTPStatus.OK, headers=None):
        if isinstance(body, (dict, list)):
            resp = web.json_response(body, status=status, headers=headers)
        elif path is not None:
            resp = self.file_response(path, status=status, headers=headers)
        else:
            resp = web.Response(body=body, status=status, headers=headers)
        resp.headers["Docker-Distribution-Api-Version"] = "registry/2.0"
        return resp

    def repo(self):
        try:
            return super().repo()
        except web.HTTPForbidden as e:
            raise errors.Denied(str(e)) from e

    def registry(self):
        try:
            return super().repo().container_registry
        except web.HTTPForbidden as e:
            raise errors.Denied(str(e)) from e


# --------------------------------------------------------------------------------------
class RootView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/"

    async def get(self):
        return self.response({})


# --------------------------------------------------------------------------------------
class CatalogView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/_catalog"

    async def get(self):
        repositories = []
        for image in self.registry().repositories():
            if self.access_granted(f"/v2/{image}/"):
                repositories.append(image)
        return self.response({"repositories": repositories})


# --------------------------------------------------------------------------------------
class ManifestReadOnlyView(ContainerView):
    CONTENT_TYPE = "application/vnd.docker.distribution.manifest.v2+json"

    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/manifests/{reference}"
        yield "/v2/u/{user}/{branch}/{job}/manifests/{reference}"
        yield "/v2/{product}/{variant}/{product_branch}/manifests/{reference}"
        yield "/v2/u/{user}/{product}/{variant}/{product_branch}/manifests/{reference}"

    def get_manifest(self):
        match = self.request.match_info
        ref = match["reference"]
        registry = self.registry()
        try:
            if ":" in ref:
                path = registry.manifest_by_digest(ref)
                digest = ref
            else:
                if "product" in match:
                    parent = (
                        self.repo()
                        .get_product(match["product"])
                        .get_variant(match["variant"])
                        .get_branch(match["product_branch"])
                        .get_version(ref, self.access_granted)
                    )
                else:
                    parent = (
                        self.repo()
                        .get_branch(match["branch"])
                        .get_tag(ref, self.access_granted)
                        .get_job(match["job"])
                    )
                path, digest = registry.manifest_by_parent(parent.path())
        except ValueError as e:
            raise errors.ManifestInvalid(ref) from e
        except FileNotFoundError as e:
            raise errors.ManifestUnknown(ref) from e

        return path, digest

    async def head(self):
        self.get_manifest()
        return self.response()

    async def get(self):
        path, digest = self.get_manifest()
        return self.response(
            path=path,
            headers={
                "Content-Type": self.CONTENT_TYPE,
                "Docker-Content-Digest": digest,
            },
        )


# --------------------------------------------------------------------------------------
class ManifestView(ManifestReadOnlyView):
    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/manifests/{reference}"
        yield "/v2/u/{user}/{branch}/{job}/manifests/{reference}"

    async def put(self):
        match = self.request.match_info
        if "product" in match:
            raise errors.Unsupported("uploading manifests by product is not supported")
        tag = match["reference"]
        if ":" in tag:
            raise errors.Unsupported("uploading manifests by digest is not supported")
        if tag in ("latest", "stable"):
            raise errors.Unsupported(
                "uploading explicit 'latest' or 'stable' tags is not supported"
            )
        job = self.repo().get_branch(match["branch"]).get_tag(tag).get_job(match["job"])
        try:
            manifest = await self.request.json()
            if not isinstance(manifest, dict):
                raise ValueError()
            if manifest.get("mediaType") != self.CONTENT_TYPE:
                raise errors.Unsupported()
            digest = self.registry().new_manifest(job, manifest)
        except PermissionError as e:
            raise errors.Denied(str(e)) from e
        except ValueError as e:
            raise errors.ManifestInvalid() from e
        except FileNotFoundError as e:
            raise errors.BlobUnknown() from e

        return self.response(
            status=HTTPStatus.CREATED,
            headers={
                "Location": self.request.path,
                "Content-Length": "0",
                "Docker-Content-Digest": digest,
            },
        )


# --------------------------------------------------------------------------------------
class TagsListView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/tags/list"
        yield "/v2/u/{user}/{branch}/{job}/tags/list"
        yield "/v2/{product}/{variant}/{product_branch}/tags/list"
        yield "/v2/u/{user}/{product}/{variant}/{product_branch}/tags/list"

    async def get(self):
        match = self.request.match_info
        try:
            if "product" in match:
                args = (match["product"], match["variant"], match["product_branch"])
                tags = self.registry().product_tags(*args)
            else:
                args = (match["branch"], match["job"])
                tags = self.registry().job_tags(*args)
        except FileNotFoundError as e:
            raise errors.ManifestUnknown(self.request.path) from e

        image = "/".join(args)
        if "user" in match:
            root_url = f"/v2/{match['user']}/{image}"
        else:
            root_url = f"/v2/{image}"

        accessible_tags = []
        for t in tags:
            if self.access_granted(f"{root_url}/manifests/{t}"):
                accessible_tags.append(t)

        return self.response({"name": image, "tags": accessible_tags})


# --------------------------------------------------------------------------------------
class NewBlobUploadView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/blobs/uploads/"
        yield "/v2/u/{user}/{branch}/{job}/blobs/uploads/"

    async def post(self):
        uuid = self.repo().next_upload()
        return self.response(
            status=HTTPStatus.ACCEPTED,
            headers={
                "Location": f"{self.request.path}{uuid}",
                "Range": "0-0",
                "Docker-Upload-Uuid": uuid,
            },
        )


# --------------------------------------------------------------------------------------
class BlobsUploadsView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/blobs/uploads/{uuid}"
        yield "/v2/u/{user}/{branch}/{job}/blobs/uploads/{uuid}"

    async def patch(self):
        uuid = self.request.match_info["uuid"]
        try:
            data_size = await self.repo().update_upload(uuid, self.request.content.read)
        except PermissionError as e:
            raise errors.Denied(str(e)) from e
        except FileNotFoundError as e:
            raise errors.BlobUploadUnknown() from e
        return self.response(
            status=HTTPStatus.NO_CONTENT,
            headers={
                "Location": self.request.path,
                "Range": f"0-{data_size - 1}",
                "Docker-Upload-Uuid": uuid,
            },
        )

    async def put(self):
        uuid = self.request.match_info["uuid"]
        try:
            digest = self.request.query["digest"]
            if int(self.request.headers.get("Content-Length", "0")) > 0:
                await self.repo().update_upload(uuid, self.request.content.read)
            await self.repo().finalize_upload(uuid, digest)
        except PermissionError as e:
            raise errors.Denied(str(e)) from e
        except ValueError as e:
            raise errors.BlobUploadInvalid() from e
        except KeyError as e:
            self.repo().cancel_upload(uuid)
            raise errors.BlobUploadInvalid() from e
        except FileNotFoundError as e:
            raise errors.BlobUploadUnknown() from e
        return self.response(
            status=HTTPStatus.NO_CONTENT,
            headers={
                "Location": self.request.path,
                "Docker-Content-Digest": digest,
            },
        )

    async def delete(self):
        try:
            self.repo().cancel_upload(self.request.match_info["uuid"])
        except FileNotFoundError as e:
            raise errors.BlobUploadUnknown() from e
        return self.response()


# --------------------------------------------------------------------------------------
class BlobsView(ContainerView):
    @classmethod
    def urls(cls):
        yield "/v2/{branch}/{job}/blobs/{digest}"
        yield "/v2/u/{user}/{branch}/{job}/blobs/{digest}"
        yield "/v2/{product}/{variant}/{product_branch}/blobs/{digest}"
        yield "/v2/u/{user}/{product}/{variant}/{product_branch}/blobs/{digest}"

    def _blob_path(self):
        digest = self.request.match_info["digest"]
        try:
            path = self.repo().blob_path(digest)
        except ValueError as e:
            raise errors.BlobUnknown(str(e)) from e
        if not path.is_file():
            raise errors.BlobUnknown()
        return path, digest

    async def head(self):
        path, digest = self._blob_path()
        return self.response(
            headers={
                "Content-Length": str(path.stat().st_size),
                "Docker-Content-Digest": digest,
            },
        )

    async def get(self):
        path, digest = self._blob_path()
        return self.response(path=path, headers={"Docker-Content-Digest": digest})



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import logging
import os
from pathlib import Path
import socket
from typing import Iterator, Optional

import aiohttp

from .fmt import ArtifactFormat
from .job import Job
from .util import SubDir


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class Tag(SubDir):
    """
    TODO
    """

    def create(self):
        super().create()
        stamp = self._path / ".stamp"
        if not stamp.exists():
            stamp.touch()

    @classmethod
    def creation_date(cls, t):
        stamp = t.path() / ".stamp"
        if stamp.is_file():
            # prefer mtime over ctime
            # on UNIX, ctime is "the time of most recent metadata change" whereas
            # mtime is "most recent content modification"
            return stamp.stat().st_mtime
        return 0

    @property
    def timestamp(self) -> int:
        return Tag.creation_date(self)

    def get_jobs(self) -> Iterator[Job]:
        yield from Job.all(self)

    def get_job(self, name: str) -> Job:
        return Job(self, name)

    def _publish_status_path(self) -> Path:
        return self._path / ".publish-status"

    def publish_status(self) -> Optional[str]:
        try:
            return self._publish_status_path().read_text().strip()
        except FileNotFoundError:
            return None

    def _released_path(self) -> Path:
        return self._path / ".released"

    def is_released(self) -> bool:
        return self._released_path().is_file()

    def set_released(self, released: bool, semaphore: asyncio.Semaphore):
        if not self._path.is_dir():
            raise FileNotFoundError()
        loop = asyncio.get_running_loop()
        task = loop.create_task(self.do_release(released, semaphore))
        task.add_done_callback(self.done_cb)

    def done_cb(self, task):
        if task.cancelled():
            return
        exc = task.exception()
        if exc:
            LOG.error("while changing released flag on tag %s", self.name, exc_info=exc)
            self._publish_status_path().write_text(f"error: {exc}\n")

    PUBLISH_URL = os.getenv("DLREPO_PUBLISH_URL")
    PUBLISH_AUTH = os.getenv("DLREPO_PUBLISH_AUTH")
    USER_AGENT = f"dlrepo-server/{socket.gethostname()}"

    def _publish_session(self) -> aiohttp.ClientSession:
        with open(self.PUBLISH_AUTH, "r", encoding="utf-8") as f:
            buf = f.read().strip()
        if ":" not in buf:
            raise ValueError("invalid DLREPO_PUBLISH_AUTH file")
        login, password = buf.split(":", 1)
        auth = aiohttp.BasicAuth(login, password, "utf-8")
        return aiohttp.ClientSession(
            self.PUBLISH_URL,
            auth=auth,
            raise_for_status=True,
            headers={"User-Agent": self.USER_AGENT},
        )

    async def do_release(self, released: bool, semaphore: asyncio.Semaphore):
        if self.PUBLISH_URL and self.PUBLISH_AUTH:
            self._publish_status_path().write_text("in progress\n")
            async with self._publish_session() as sess:
                if released:
                    LOG.info(
                        "publishing tag %s/%s to %s",
                        self.parent.name,
                        self.name,
                        self.PUBLISH_URL,
                    )
                    await self._publish(sess, semaphore)
                else:
                    LOG.info(
                        "deleting tag %s/%s from %s",
                        self.parent.name,
                        self.name,
                        self.PUBLISH_URL,
                    )
                    async with semaphore:
                        await sess.delete(self.url(), params={"force": "true"})
                    p = self._publish_status_path()
                    if p.is_file():
                        p.unlink()
        path = self._released_path()
        if released:
            path.touch()
        elif path.is_file():
            path.unlink()

    async def _publish(self, sess: aiohttp.ClientSession, semaphore: asyncio.Semaphore):
        loop = asyncio.get_running_loop()
        for job in self.get_jobs():
            if job.is_internal():
                continue
            job_url = job.url()
            # in case job is already present on the server, delete it before re-uploading it
            await sess.delete(job_url, params={"force": "true"}, raise_for_status=False)
            self._publish_status_path().write_text(f"uploading {job.name}\n")
            tasks = []
            for fmt in job.get_formats():
                tasks.append(loop.create_task(self._publish_fmt(fmt, sess, semaphore)))
            await asyncio.gather(*tasks)
            metadata = job.get_metadata()
            del metadata["name"]
            del metadata["locked"]
            LOG.debug("publishing job metadata %s", job_url)
            async with semaphore:
                await sess.patch(job_url, json={"job": metadata})
        self._publish_status_path().write_text(f"published to {self.PUBLISH_URL}\n")

    async def _publish_fmt(
        self,
        fmt: ArtifactFormat,
        sess: aiohttp.ClientSession,
        semaphore: asyncio.Semaphore,
    ):
        fmt_url = fmt.url()
        loop = asyncio.get_running_loop()

        async def _publish_file(file, digest):
            file_url = fmt_url + file
            headers = {"Digest": digest}

            async with semaphore:
                resp = await sess.head(
                    file_url, headers=headers, raise_for_status=False
                )
            if resp.status == 200:
                LOG.debug("publishing file %s (deduplicated)", file_url)
                # file digest already present on the server, do not upload
                # the data again
                headers["X-Dlrepo-Link"] = digest
                async with semaphore:
                    await sess.put(file_url, data=None, headers=headers)

            else:
                LOG.debug("publishing file %s", file_url)
                # file digest not on server, proceed with upload
                async with semaphore:
                    with open(fmt.path() / file, "rb") as f:
                        await sess.put(file_url, data=f, headers=headers)

        tasks = []
        for file, digest in fmt.get_digests().items():
            tasks.append(loop.create_task(_publish_file(file, digest)))
        await asyncio.gather(*tasks)

        # clear the dirty flag
        async with semaphore:
            await sess.patch(fmt_url)

    def _locked_path(self) -> Path:
        return self._path / ".locked"

    def is_locked(self) -> bool:
        return self._locked_path().is_file()

    def set_locked(self, locked: bool):
        path = self._locked_path()
        if locked:
            path.touch()
        elif path.is_file():
            path.unlink()

    def delete(self, *, force: bool = False):
        if not self.exists():
            raise FileNotFoundError()
        if self.is_locked():
            raise OSError(f"Tag {self.name} is locked")
        if not force and self.is_released():
            raise OSError(f"Tag {self.name} is released, use force")
        for j in self.get_jobs():
            j.delete()
        self.root().rmtree(self._path)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import os

from aiohttp import web

from .util import BaseView


# --------------------------------------------------------------------------------------
class ArtifactView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}/{format}/{artifact:.+}"
        yield "/~{user}/branches/{branch}/{tag}/{job}/{format}/{artifact:.+}"
        yield "/%7e{user}/branches/{branch}/{tag}/{job}/{format}/{artifact:.+}"
        yield "/%7E{user}/branches/{branch}/{tag}/{job}/{format}/{artifact:.+}"
        yield "/products/{product}/{variant}/{product_branch}/{version}/{format}/{artifact:.+}"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/{format}/{artifact:.+}"
        yield "/%7e{user}/products/{product}/{variant}/{product_branch}/{version}/{format}/{artifact:.+}"
        yield "/%7E{user}/products/{product}/{variant}/{product_branch}/{version}/{format}/{artifact:.+}"

    def _get_format(self, create=False):
        match = self.request.match_info
        if "product" in match:
            if create:
                raise web.HTTPBadRequest(reason="Uploading products is not supported")
            fmt = (
                self.repo()
                .get_product(match["product"])
                .get_variant(match["variant"])
                .get_branch(match["product_branch"])
                .get_version(match["version"])
                .get_format(match["format"])
            )
        else:
            job = (
                self.repo()
                .get_branch(match["branch"])
                .get_tag(match["tag"])
                .get_job(match["job"])
            )
            fmt = job.get_format(match["format"])
            if create:
                if match["tag"] == "latest":
                    raise web.HTTPMethodNotAllowed(self.request.method, ["GET"])
                if match["format"] == "container":
                    agent = self.request.headers.get("User-Agent", "")
                    # allow uploading directly between dlrepo servers
                    if not agent.startswith("dlrepo-server/"):
                        raise web.HTTPBadRequest(
                            reason="Uploading container images must be done with docker push"
                        )
                if job.is_locked():
                    raise web.HTTPBadRequest(
                        reason="Cannot upload files in locked jobs"
                    )
        if not create and not fmt.exists():
            raise web.HTTPNotFound()
        return fmt

    async def head(self):
        digest = self.request.headers.get("Digest")
        if digest is not None:
            try:
                filepath = self.repo().blob_path(digest)
                if not filepath.is_file():
                    raise web.HTTPNotFound()
            except ValueError as e:
                raise web.HTTPUnprocessableEntity(reason=str(e)) from e

        else:
            fmt = self._get_format()
            artifact = self.request.match_info["artifact"]
            try:
                filepath = fmt.get_filepath(artifact)
            except PermissionError as e:
                raise web.HTTPBadRequest(reason=str(e)) from e
            except FileNotFoundError as e:
                raise web.HTTPNotFound() from e
            if filepath.is_dir():
                raise web.HTTPNotFound()
            url = fmt.url() + artifact
            if url != self.request.path:
                raise web.HTTPFound(url)
            digest = fmt.get_digests()[artifact]

        return web.Response(
            headers={"Digest": digest, "Content-Length": str(filepath.stat().st_size)}
        )

    async def get(self):
        fmt = self._get_format()
        artifact = self.request.match_info["artifact"]
        try:
            filepath = fmt.get_filepath(artifact)
        except PermissionError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e
        except FileNotFoundError as e:
            if artifact.endswith(".sha256"):
                artifact = artifact[: -len(".sha256")]
                digest = fmt.get_digests().get(artifact)
                if digest is not None:
                    algo, digest = digest.split(":")
                    if algo == "sha256":
                        url = fmt.url() + artifact + ".sha256"
                        if url != self.request.path:
                            raise web.HTTPFound(url)
                        artifact = os.path.basename(artifact)
                        return web.Response(text=f"{digest}  {artifact}\n")
            raise web.HTTPNotFound() from e
        url = fmt.url() + artifact
        if filepath.is_dir():
            if "html" in self.request.headers.get("Accept", "json"):
                if not url.endswith("/"):
                    raise web.HTTPFound(url + "/")
                if url != self.request.path:
                    raise web.HTTPFound(url)
                return await self.autoindex(fmt, artifact)
            raise web.HTTPNotFound()
        if url != self.request.path:
            raise web.HTTPFound(url)
        digest = fmt.get_digests().get(artifact)
        if not digest and fmt.name == "container":
            algo = filepath.parent.parent.name
            if algo in hashlib.algorithms_guaranteed:
                digest = f"{algo}:{filepath.name}"
        headers = {}
        if digest:
            headers["Digest"] = digest
        return self.file_response(filepath, headers=headers)

    async def put(self):
        fmt = self._get_format(create=True)
        artifact = self.request.match_info["artifact"]
        digest = self.request.headers.get("Digest")
        try:
            if self.request.headers.get("X-Dlrepo-Link") == digest:
                # no content, link existing blob
                fmt.link_file(digest, artifact)
            else:
                await fmt.add_file(artifact, self.request.content.read, digest)
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e
        except PermissionError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e
        except ValueError as e:
            raise web.HTTPUnprocessableEntity(reason=str(e)) from e
        return web.HTTPCreated()



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import errno
import json
import logging
import os
from pathlib import Path
from typing import Dict, Iterator

from .fmt import ArtifactFormat
from .product import Version
from .util import SubDir


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class Job(SubDir):
    """
    TODO
    """

    def create(self):
        super().create()
        stamp = self._path / ".stamp"
        if not stamp.exists():
            stamp.touch()

    @classmethod
    def creation_date(cls, j):
        stamp = j.path() / ".stamp"
        if stamp.is_file():
            # prefer mtime over ctime
            # on UNIX, ctime is "the time of most recent metadata change" whereas
            # mtime is "most recent content modification"
            return stamp.stat().st_mtime
        return 0

    @property
    def timestamp(self) -> int:
        return Job.creation_date(self)

    def get_formats(self) -> Iterator[ArtifactFormat]:
        yield from ArtifactFormat.all(self)

    def get_format(self, name: str) -> ArtifactFormat:
        return ArtifactFormat(self, name)

    def archive_name(self) -> str:
        data = self.get_metadata()
        if {"product", "product_variant", "version"} <= set(data):
            return f"{data['product']}-{data['product_variant']}-v{data['version']}"
        return f"{self.name}-{self.parent.name}"

    def _metadata_path(self):
        return self._path / ".metadata"

    def _product_link_path(self):
        return self._path / ".product"

    @staticmethod
    def update_symlink(dst, link):
        if link.exists():
            if link.is_symlink():
                link.unlink()
            else:
                raise OSError(errno.ENOTEMPTY, f"{link} exists and is not a symlink")
        link.symlink_to(os.path.relpath(dst, link.parent))

    def _link_to_product(self, version: Version):
        """
        Link the job to its product, according to the job metadata.
        Create the product fs if needed.
        The links are created in both ways:

        - From the job to its product version:
            $ROOT/branches/<branch>/<tag>/<job>/.product ->
                $ROOT/products/<product>/<variant>/<branch>/<version>

        - From the product version to the job(s), for each format:
            $ROOT/products/<product>/<variant>/<branch>/<version>/<format> ->
                $ROOT/branches/<branch>/<tag>/<job>/<format>

        The formats in a product version are not necessarily linked to the same job.
        """
        version.create()
        self.update_symlink(version.path(), self._product_link_path())
        for fmt in self.get_formats():
            self.update_symlink(fmt.path(), version.path() / fmt.name)
        container = self.path() / "container"
        if container.is_dir():
            self.update_symlink(container, version.path() / "container")

    def _cleanup_product_tree(self):
        link = self._product_link_path()
        if not link.is_symlink():
            return
        if not link.is_dir():
            link.unlink()
            return
        product = link.resolve()
        for d in product.iterdir():
            if not d.is_symlink():
                continue
            if not d.is_dir():
                continue
            try:
                if d.resolve().samefile(self.path() / d.name):
                    d.unlink()
            except FileNotFoundError:
                # same product, different jobs (e.g. doc + binaries)
                pass
        try:
            if os.listdir(product) == [".stamp"]:
                (product / ".stamp").unlink()
            # cleanup empty dirs
            os.removedirs(product)
        except OSError:
            # directory not empty, abort
            pass
        link.unlink()

    def get_metadata(self) -> Dict:
        try:
            data = json.loads(self._metadata_path().read_text())
        except (OSError, ValueError):
            data = {}
        data["locked"] = self.is_locked()
        data["name"] = self.name
        return data

    def _lock_path(self):
        return self.path() / ".locked"

    def is_locked(self) -> bool:
        return self._lock_path().is_file()

    def set_locked(self, locked: bool):
        if not self.exists():
            raise FileNotFoundError()
        self.create()
        path = self._lock_path()
        if locked:
            path.touch()
        elif path.is_file():
            path.unlink()

    def _internal_path(self) -> Path:
        return self._path / ".internal"

    def is_internal(self) -> bool:
        return self._internal_path().is_file()

    def set_internal(self, internal: bool):
        path = self._internal_path()
        if internal:
            path.touch()
        elif path.is_file():
            path.unlink()

    def add_metadata(self, new_data: Dict):
        if self.is_locked():
            raise FileExistsError("Job is locked")
        self._cleanup_product_tree()
        metadata_path = self._metadata_path()
        try:
            data = json.loads(metadata_path.read_text())
        except (OSError, ValueError):
            data = {}
        for k, v in new_data.items():
            v = str(v or "").lower()
            if v == "":
                data.pop(k, None)
            else:
                data[k] = v
        self.create()
        metadata_path.write_text(json.dumps(data))
        if {"product", "version", "product_branch", "product_variant"} <= set(data):
            self._link_to_product(
                self.root()
                .get_product(str(data["product"]))
                .get_variant(str(data["product_variant"]))
                .get_branch(str(data["product_branch"]))
                .get_version(str(data["version"]))
            )

    def delete(self):
        if not self.exists():
            raise FileNotFoundError()
        self._cleanup_product_tree()
        self.root().rmtree(self._path)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
from http import HTTPStatus
import json
import os
from pathlib import Path
import tarfile
from typing import Dict, Tuple

from aiohttp import web
from aiohttp.abc import AbstractStreamWriter
import aiohttp_jinja2

from . import auth


# --------------------------------------------------------------------------------------
class BaseView(web.View):
    @classmethod
    def urls(cls):
        raise NotImplementedError()

    def repo(self):
        r = self.request.app["dlrepo_artifact_repository"]
        if "user" in self.request.match_info:
            r = r.get_user_repo(self.request.match_info["user"])
        return r

    def access_granted(self, url: str) -> bool:
        if auth.AuthBackend.AUTH_DISABLED:
            return True
        auth_backend = self.request.app[auth.AuthBackend.KEY]
        acls = self.request["dlrepo_user_acls"]
        return auth_backend.access_granted(acls, False, url)

    X_SENDFILE_HEADER = os.getenv("DLREPO_X_SENDFILE_HEADER", None)

    def file_response(self, path, status=HTTPStatus.OK, headers=None):
        if self.X_SENDFILE_HEADER:
            if headers is None:
                headers = {}
            headers[self.X_SENDFILE_HEADER] = str(path)
            if "Content-Type" not in headers:
                # Force empty Content-Type to prevent aiohttp from adding the
                # application/octet-stream default value.
                # The reverse proxy will guess the content type based on the file name.
                headers["Content-Type"] = ""
            return web.Response(status=status, headers=headers)
        return web.FileResponse(path, status=status, headers=headers)

    async def json_body(self):
        try:
            return await self.request.json()
        except json.JSONDecodeError as e:
            raise web.HTTPBadRequest(reason=f"invalid JSON data: {e}") from e

    async def autoindex(self, fmt, relpath):
        dirs, files = fmt.list_dir(relpath)
        if "index.html" in files:
            return self.file_response(
                fmt.get_filepath(os.path.join(relpath, "index.html"))
            )
        data = {
            "artifact_format": {
                "name": fmt.name,
                "relpath": relpath.rstrip("/"),
                "dirs": dirs,
                "files": files,
            },
        }
        return aiohttp_jinja2.render_template("autoindex.html", self.request, data)


# --------------------------------------------------------------------------------------
class TarResponse(web.StreamResponse):
    def __init__(self, members: Dict[str, str], root_dir: Path, tar_prefix: str = None):
        self._members = members
        self._root_dir = root_dir
        self._tar_prefix = tar_prefix
        super().__init__()
        self.content_type = "application/x-tar"
        if self._tar_prefix is not None:
            filename = self._tar_prefix + ".tar"
            self.headers.add("Content-Disposition", f"attachment; filename={filename}")

    async def prepare(self, request: web.Request) -> AbstractStreamWriter:
        # pylint: disable=too-many-locals
        writer = await super().prepare(request)
        transport = request.transport
        loop = asyncio.get_running_loop()

        tar_offset = 0
        sha256sums = []

        for member, digest in sorted(self._members.items()):
            info, path = await self._get_tar_info(member)
            algo, digest = digest.split(":")
            if algo == "sha256":
                sha256sums.append(f"{digest}  {member}\n")

            buf = info.tobuf(tarfile.PAX_FORMAT, "utf-8")
            await self.write(buf)  # includes http chunk header & trailer
            tar_offset += len(buf)

            fobj = await loop.run_in_executor(None, path.open, "rb")
            try:
                # loop.sendfile works on the raw transport class and does not include
                # http chunk headers/trailers. Manually insert them.
                transport.write(self.CHUNK_HEADER % info.size)
                await loop.sendfile(transport, fobj, count=info.size)
                transport.write(self.CHUNK_TRAILER)

                blocks, remainder = divmod(info.size, tarfile.BLOCKSIZE)
                if remainder > 0:
                    # add tar member padding (includes http chunk header & trailer)
                    await self.write(self._padding(tarfile.BLOCKSIZE - remainder))
                    blocks += 1
                tar_offset += blocks * tarfile.BLOCKSIZE
            finally:
                await loop.run_in_executor(None, fobj.close)

        if sha256sums:
            tar_offset += await self._write_sha256sums(sha256sums)

        # add tar format trailer
        tar_trailer = self._padding(2 * tarfile.BLOCKSIZE)
        tar_offset += 2 * tarfile.BLOCKSIZE
        remainder = tar_offset % tarfile.RECORDSIZE
        if remainder > 0:
            tar_trailer += self._padding(tarfile.RECORDSIZE - remainder)

        await self.write_eof(tar_trailer)  # includes http chunk header & trailer

        return writer

    CHUNK_HEADER = b"%x\r\n"
    CHUNK_TRAILER = b"\r\n"

    @staticmethod
    def _padding(size):
        return b"\x00" * size

    async def _get_tar_info(self, name: str) -> Tuple[tarfile.TarInfo, Path]:
        path = self._root_dir / name
        if self._tar_prefix is not None:
            name = f"{self._tar_prefix}/{name}"

        loop = asyncio.get_running_loop()
        st = await loop.run_in_executor(None, path.lstat)

        info = tarfile.TarInfo(name)
        info.mode = st.st_mode
        info.size = st.st_size

        return info, path

    async def _write_sha256sums(self, sha256sums):
        name = "SHA256SUMS.txt"
        if self._tar_prefix is not None:
            name = f"{self._tar_prefix}/{name}"

        data = "".join(sha256sums).encode("utf-8")

        info = tarfile.TarInfo(name)
        info.mode = 0o0644
        info.size = len(data)

        buf = info.tobuf(tarfile.PAX_FORMAT, "utf-8") + data
        offset = len(buf)
        blocks, remainder = divmod(info.size, tarfile.BLOCKSIZE)
        if remainder > 0:
            # add tar member padding
            buf += self._padding(tarfile.BLOCKSIZE - remainder)
            blocks += 1
        offset += blocks * tarfile.BLOCKSIZE

        await self.write(buf)  # includes http chunk header & trailer

        return offset



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import json
import logging
import os
from pathlib import Path
import re
from typing import Awaitable, Callable, Dict, Iterator, List, Tuple

from cachetools import LRUCache, cachedmethod

from .util import SubDir, file_digest


LOG = logging.getLogger(__name__)


# --------------------------------------------------------------------------------------
class ArtifactFormat(SubDir):
    """
    TODO
    """

    def url_bit(self) -> str:
        return self.name

    def get_files(self) -> Iterator[str]:
        for root, dirs, files in os.walk(self._path):
            dirs.sort()
            files.sort()
            for f in files:
                f = Path(root, f)
                if self._is_reserved_file(f):
                    continue
                if f.is_file():
                    yield str(f.relative_to(self._path))

    def archive_name(self) -> str:
        return f"{self.parent.archive_name()}-{self.name}"

    _is_reserved_cache = LRUCache(4096)

    @cachedmethod(lambda self: self._is_reserved_cache)
    def _is_reserved_file(self, path, *, resolve=False):
        digests = self._digest_path()
        dirty = self._dirty_path()
        if resolve:
            digests = digests.resolve()
            dirty = dirty.resolve()
        return path in (digests, dirty)

    def list_dir(self, relpath: str) -> Tuple[List[str], List[str]]:
        path = self.get_filepath(relpath)
        if not path.is_dir():
            raise NotADirectoryError(relpath)
        dirs = []
        files = []
        for e in path.iterdir():
            if self._is_reserved_file(e, resolve=True):
                continue
            if e.is_dir():
                dirs.append(e.name)
            elif e.is_file():
                files.append(e.name)
        return dirs, files

    def _check_filepath(self, relpath: str) -> Path:
        if relpath.startswith("/") or any(x in (".", "..") for x in relpath.split("/")):
            raise PermissionError(relpath)
        path = self._path / relpath
        if self._is_reserved_file(path):
            raise PermissionError(relpath)
        return path

    def get_filepath(self, relpath: str) -> Path:
        return self._check_filepath(relpath).resolve(strict=True)

    def get_digests(self) -> Dict[str, str]:
        try:
            return json.loads(self._digest_path().read_text())
        except FileNotFoundError:
            if self.name != "container":
                return {}
            try:
                return {f: f"sha256:{os.path.basename(f)}" for f in self.get_files()}
            except OSError:
                return {}
        except (OSError, ValueError):
            return {}

    def _digest_path(self) -> Path:
        return self._path / ".digests"

    async def add_file(
        self,
        relpath: str,
        read: Callable[[int], Awaitable[bytes]],
        digest: str,
    ):
        self._check_filepath(relpath)
        uuid = self.root().next_upload()
        await self.root().update_upload(uuid, read)
        await self.root().finalize_upload(uuid, digest)
        # avoid counting disk usage twice (already counted in update_upload()
        self.link_file(digest, relpath, ignore_quota=True)

    def link_file(self, digest: str, relpath: str, ignore_quota: bool = False):
        was_dirty = self.is_dirty()
        self.set_dirty(True)
        try:
            path = self._check_filepath(relpath)
            if ignore_quota:
                self.root().link_blob_ignore_quota(digest, path)
            else:
                self.root().link_blob(digest, path)
        except:
            if not was_dirty:
                self.set_dirty(False)
            raise
        # update digests file
        digests = self.get_digests()
        digests[relpath] = digest
        self._digest_path().write_text(json.dumps(digests))

    def _dirty_path(self) -> Path:
        return self._path / ".dirty"

    def is_dirty(self) -> bool:
        return self._dirty_path().is_file()

    def set_dirty(self, dirty: bool):
        path = self._dirty_path()
        if dirty:
            self.create()
            path.touch()
        elif path.is_file():
            path.unlink()
            try:
                os.removedirs(path.parent)
            except OSError:
                pass

    POST_PROCESS_CMD = os.getenv("DLREPO_POST_PROCESS_CMD")
    POST_PROCESS_FILTER = os.getenv("DLREPO_POST_PROCESS_FILTER")

    async def post_process(self):
        if not self.POST_PROCESS_CMD:
            return

        digests = self.get_digests()
        if self.POST_PROCESS_FILTER and not any(
            re.match(self.POST_PROCESS_FILTER, f) for f in digests.keys()
        ):
            return

        LOG.debug(
            "running post process command: %s (cwd %s)",
            self.POST_PROCESS_CMD,
            self._path,
        )

        ctimes = {}
        for f in digests.keys():
            # make a temporary copy of all hardlinked files so that post-process can
            # safely modify them
            self.root().copy_blob(digests[f], self._path / f)
            # record content modification time to only recalculate digest if needed
            ctimes[f] = (self._path / f).stat().st_ctime

        try:
            proc = await asyncio.create_subprocess_exec(
                self.POST_PROCESS_CMD, cwd=self._path
            )
            ret = await proc.wait()
            if ret != 0:
                raise OSError(
                    f"command {self.POST_PROCESS_CMD} exited {ret} (cwd {self._path})"
                )
        finally:
            # always refresh digests even on error to ensure consistency
            await self._refresh_digests(digests, ctimes)

    async def _refresh_digests(self, digests, ctimes):
        new_digests = {}
        loop = asyncio.get_running_loop()

        for f in self.get_files():
            fpath = self._path / f
            if f in ctimes and f in digests and ctimes[f] == fpath.stat().st_ctime:
                # file was not modified by post processing, avoid recalculating digest
                digest = new_digests[f] = digests[f]
            else:
                # new or modified file
                digest = await loop.run_in_executor(None, file_digest, "sha256", fpath)
                digest = f"sha256:{digest}"
                new_digests[f] = digest

            # deduplicate files, remove temporary copies
            blob = self.root().blob_path(digest)
            if blob.is_file():
                # replace file with hardlink to existing blob
                fpath.unlink()
                os.link(blob, fpath)
            else:
                # add hardlink to new blob
                blob.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
                os.link(fpath, blob)

        self._digest_path().write_text(json.dumps(new_digests))



# Copyright (c) 2022 Julien Floret
# Copyright (c) 2022 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import hashlib

import aiohttp
import pytest


pytestmark = pytest.mark.asyncio
X_SENDFILE_HEADER = "X-Sendfile-Foobar"


async def test_x_sendfile(dlrepo_server):
    url, data_dir = dlrepo_server
    async with aiohttp.ClientSession(url) as sess:
        data2 = b"content2"
        digest2 = hashlib.sha256(data2).hexdigest()
        resp = await sess.put(
            "/branches/main/tag/job/format/file2.txt",
            data=data2,
            headers={"Digest": f"sha256:{digest2}"},
        )
        assert resp.status == 201

        data = b"content"
        digest = hashlib.sha256(data).hexdigest()
        resp = await sess.put(
            "/branches/main/tag/job/format/file.txt",
            data=data,
            headers={"Digest": f"sha256:{digest}"},
        )
        assert resp.status == 201

        for f in ("file.txt", "file2.txt"):
            resp = await sess.get(f"/branches/main/tag/job/format/{f}")
            assert resp.status == 200
            assert X_SENDFILE_HEADER in resp.headers
            path = data_dir / "branches/main/tag/job/format" / f
            assert resp.headers[X_SENDFILE_HEADER] == str(path)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio

from aiohttp import web
import aiohttp_jinja2

from .util import BaseView, TarResponse


# --------------------------------------------------------------------------------------
class JobArchiveView(BaseView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}.tar"
        yield "/~{user}/branches/{branch}/{tag}/{job}.tar"

    def _get_job(self):
        match = self.request.match_info
        if match["tag"] in ("latest", "stable") and self.request.method != "GET":
            raise web.HTTPMethodNotAllowed(self.request.method, ["GET"])
        try:
            return (
                self.repo()
                .get_branch(match["branch"])
                .get_tag(match["tag"])
                .get_job(match["job"])
            )
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e

    def _digests(self, job):
        digests = {}
        for fmt in job.get_formats():
            if self.access_granted(fmt.url()):
                for f, digest in fmt.get_digests().items():
                    digests[f"{fmt.name}/{f}"] = digest
        return digests

    async def get(self):
        job = self._get_job()
        url = job.url().rstrip("/") + ".tar"
        if url != self.request.path:
            raise web.HTTPFound(url)
        return TarResponse(self._digests(job), job.path(), job.archive_name())


# --------------------------------------------------------------------------------------
class JobView(JobArchiveView):
    @classmethod
    def urls(cls):
        yield "/branches/{branch}/{tag}/{job}"
        yield "/branches/{branch}/{tag}/{job}/"
        yield "/~{user}/branches/{branch}/{tag}/{job}"
        yield "/~{user}/branches/{branch}/{tag}/{job}/"

    async def get(self):
        """
        Get info about a job including metadata and artifact formats.
        """
        job = self._get_job()
        if not job.exists():
            raise web.HTTPNotFound()
        if job.url() != self.request.path:
            raise web.HTTPFound(job.url())
        html = "html" in self.request.headers.get("Accept", "json")
        data = {"job": job.get_metadata()}
        data["job"]["internal"] = job.is_internal()
        data["job"]["timestamp"] = job.timestamp
        formats = []
        for f in job.get_formats():
            fmt_url = f.url()
            if self.access_granted(fmt_url):
                if html:
                    digests = f.get_digests()
                    deb = rpm = False
                    if "repodata/repomd.xml" in digests:
                        rpm = True
                    elif "Release" in digests:
                        deb = True
                    formats.append(
                        {
                            "name": f.name,
                            "rpm": rpm,
                            "deb": deb,
                            "url": fmt_url,
                        }
                    )
                else:
                    formats.append(f.name)
        data["job"]["artifact_formats"] = formats
        if html:
            return aiohttp_jinja2.render_template("job.html", self.request, data)
        return web.json_response(data)

    async def put(self):
        job = self._get_job()
        try:
            data = (await self.json_body())["job"]
            internal = data.get("internal")
            if internal is not None and not isinstance(internal, bool):
                raise TypeError()
            locked = data.get("locked")
            if locked is not None and not isinstance(locked, bool):
                raise TypeError()
        except (TypeError, KeyError) as e:
            raise web.HTTPBadRequest(reason="Invalid parameters") from e

        try:
            if internal is not None:
                job.set_internal(internal)
            if locked is not None:
                job.set_locked(locked)
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e

        return web.Response()

    async def patch(self):
        """
        Update the metadata of a job.
        """
        try:
            data = (await self.json_body())["job"]
            if not isinstance(data, dict):
                raise TypeError()
        except (TypeError, KeyError) as e:
            raise web.HTTPBadRequest(reason="Invalid parameters") from e
        try:
            self._get_job().add_metadata(data)
        except FileExistsError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e
        return web.Response()

    async def delete(self):
        """
        Delete a job and all its contents.
        """
        loop = asyncio.get_running_loop()
        try:
            job = self._get_job()
            await loop.run_in_executor(None, job.delete)
            self.repo().schedule_cleanup_orphans()
        except FileNotFoundError as e:
            raise web.HTTPNotFound() from e
        except OSError as e:
            raise web.HTTPBadRequest(reason=str(e)) from e
        return web.Response()



#!/usr/bin/env python3
# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import os
import pathlib

import setuptools
import setuptools.command.build_py


class BuildPyCommand(setuptools.command.build_py.build_py):
    def build_package_data(self):
        super().build_package_data()
        src = "scss/main.scss"
        if self.distribution.src_root is not None:
            src = os.path.join(self.distribution.src_root, src)
        dest = os.path.join(self.build_lib, "dlrepo/static/dlrepo.css")
        self.spawn(["sassc", "-t", "compact", src, dest])


setuptools.setup(
    name="dlrepo",
    description="Artifact repository",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text("utf-8"),
    long_description_content_type="text/markdown",
    license="BSD-3-Clause",
    version="0.24",
    author="Robin Jarry",
    author_email="robin@jarry.cc",
    install_requires=[
        "aiohttp>=3.8.0",
        "aiohttp-jinja2",
        "cachetools",
        "bonsai",
    ],
    cmdclass={
        "build_py": BuildPyCommand,
    },
    url="https://sr.ht/~rjarry/dlrepo/",
    packages=setuptools.find_packages("."),
    include_package_data=True,
    zip_safe=False,
    entry_points="""
    [console_scripts]
    dlrepo = dlrepo.__main__:main
    """,
    scripts=["dlrepo-cli"],
)



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

from aiohttp import web
import aiohttp_jinja2

from .util import BaseView, TarResponse


# --------------------------------------------------------------------------------------
class ProductsView(BaseView):
    @classmethod
    def urls(cls):
        yield "/products"
        yield "/products/"
        yield "/~{user}/products"
        yield "/~{user}/products/"

    async def get(self):
        """
        Get the list of products.
        """
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        products = []
        for p in self.repo().get_products():
            if self.access_granted(p.url()):
                products.append({"name": p.name})
        data = {"products": products}
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template("products.html", self.request, data)
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class ProductView(BaseView):
    @classmethod
    def urls(cls):
        yield "/products/{product}"
        yield "/products/{product}/"
        yield "/~{user}/products/{product}"
        yield "/~{user}/products/{product}/"

    def _get_product(self):
        product = self.repo().get_product(self.request.match_info["product"])
        if not product.exists():
            raise web.HTTPNotFound()
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        return product

    async def get(self):
        """
        Get the list of variants for a product.
        """
        product = self._get_product()
        variants = []
        for v in product.get_variants():
            if self.access_granted(v.url()):
                variants.append({"name": v.name})
        data = {"product": {"name": product.name, "product_variants": variants}}
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template("product.html", self.request, data)
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class ProductVariantView(BaseView):
    @classmethod
    def urls(cls):
        yield "/products/{product}/{variant}"
        yield "/products/{product}/{variant}/"
        yield "/~{user}/products/{product}/{variant}"
        yield "/~{user}/products/{product}/{variant}/"

    def _get_variant(self):
        variant = (
            self.repo()
            .get_product(self.request.match_info["product"])
            .get_variant(self.request.match_info["variant"])
        )
        if not variant.exists():
            raise web.HTTPNotFound()
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        return variant

    async def get(self):
        """
        Get the list of branches for a product variant.
        """
        variant = self._get_variant()
        branches = []
        for b in variant.get_branches():
            if self.access_granted(b.url()):
                branches.append({"name": b.name})
        data = {
            "product_variant": {
                "name": variant.name,
                "product_branches": branches,
            },
        }
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template(
                "product_variant.html", self.request, data
            )
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class ProductBranchView(BaseView):
    @classmethod
    def urls(cls):
        yield "/products/{product}/{variant}/{product_branch}"
        yield "/products/{product}/{variant}/{product_branch}/"
        yield "/~{user}/products/{product}/{variant}/{product_branch}"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/"

    def _get_branch(self):
        branch = (
            self.repo()
            .get_product(self.request.match_info["product"])
            .get_variant(self.request.match_info["variant"])
            .get_branch(self.request.match_info["product_branch"])
        )
        if not branch.exists():
            raise web.HTTPNotFound()
        if not self.request.path.endswith("/"):
            raise web.HTTPFound(self.request.path + "/")
        return branch

    async def get(self):
        """
        Get the list of versions for a product branch.
        """
        branch = self._get_branch()
        versions = []
        for v in branch.get_versions():
            if self.access_granted(v.url()):
                versions.append(
                    {
                        "name": v.name,
                        "timestamp": v.timestamp,
                        "locked": v.is_locked(),
                        "released": v.is_released(),
                    }
                )
        data = {
            "product_branch": {
                "name": branch.name,
                "versions": versions,
            },
        }
        if "html" in self.request.headers.get("Accept", "json"):
            return aiohttp_jinja2.render_template(
                "product_branch.html", self.request, data
            )
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class VersionView(BaseView):
    @classmethod
    def urls(cls):
        yield "/products/{product}/{variant}/{product_branch}/{version}"
        yield "/products/{product}/{variant}/{product_branch}/{version}/"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}/"

    def _get_version(self):
        version = (
            self.repo()
            .get_product(self.request.match_info["product"])
            .get_variant(self.request.match_info["variant"])
            .get_branch(self.request.match_info["product_branch"])
            .get_version(self.request.match_info["version"])
        )
        if not version.exists():
            raise web.HTTPNotFound()
        return version

    async def get(self):
        """
        Get the list of versions for a product version.
        """
        version = self._get_version()
        if version.url() != self.request.path:
            raise web.HTTPFound(version.url())
        html = "html" in self.request.headers.get("Accept", "json")

        formats = []
        for fmt in version.get_formats():
            fmt_url = fmt.url()
            if self.access_granted(fmt_url):
                if html:
                    digests = fmt.get_digests()
                    deb = rpm = False
                    if "repodata/repomd.xml" in digests:
                        rpm = True
                    elif "Release" in digests:
                        deb = True
                    formats.append(
                        {
                            "name": fmt.name,
                            "rpm": rpm,
                            "deb": deb,
                            "url": fmt_url,
                        }
                    )
                else:
                    formats.append(fmt.name)

        data = {"version": {"name": version.name, "artifact_formats": formats}}

        if html:
            return aiohttp_jinja2.render_template(
                "product_version.html", self.request, data
            )
        return web.json_response(data)


# --------------------------------------------------------------------------------------
class VersionArchiveView(VersionView):
    @classmethod
    def urls(cls):
        yield "/products/{product}/{variant}/{product_branch}/{version}.tar"
        yield "/~{user}/products/{product}/{variant}/{product_branch}/{version}.tar"

    def _digests(self, version):
        digests = {}
        for fmt in version.get_formats():
            if self.access_granted(fmt.url()):
                for f, digest in fmt.get_digests().items():
                    digests[f"{fmt.name}/{f}"] = digest
        return digests

    async def get(self):
        version = self._get_version()
        url = version.url().rstrip("/") + ".tar"
        if url != self.request.path:
            raise web.HTTPFound(version.url())
        return TarResponse(
            self._digests(version), version.path(), version.archive_name()
        )



# Copyright (c) 2021 Julien Floret
# Copyright (c) 2021 Robin Jarry
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
from datetime import datetime, timezone
import os
import pathlib

from aiohttp import web
import aiohttp_jinja2
import jinja2

from ..fs.util import human_readable
from .artifact import ArtifactView
from .branch import BranchesView, BranchView
from .container import (
    BlobsUploadsView,
    BlobsView,
    CatalogView,
    ManifestReadOnlyView,
    ManifestView,
    NewBlobUploadView,
    RootView,
    TagsListView,
)
from .fmt import FormatArchiveView, FormatDigestsView, FormatDirView, FormatFileView
from .job import JobArchiveView, JobView
from .product import (
    ProductBranchView,
    ProductsView,
    ProductVariantView,
    ProductView,
    VersionArchiveView,
    VersionView,
)
from .tag import TagView
from .util import BaseView


# --------------------------------------------------------------------------------------
class HomeView(BaseView):
    @classmethod
    def urls(cls):
        yield "/"
        yield "/~{user}"
        yield "/~{user}/"

    def _get_version_dirs(self, repo):
        stamps = list(repo.path().glob("products/*/*/*/*/.stamp"))
        # prefer mtime over ctime
        # on UNIX, ctime is "the time of most recent metadata change" whereas
        # mtime is "most recent content modification"
        stamps.sort(key=lambda s: s.stat().st_mtime, reverse=True)
        versions = []
        for s in stamps:
            versions.append(s.parent)
        return versions

    async def get_latest_releases(self, repo, num=10):
        loop = asyncio.get_running_loop()
        version_dirs = await loop.run_in_executor(None, self._get_version_dirs, repo)
        versions = []
        for v in version_dirs:
            version = (
                repo.get_product(v.parent.parent.parent.name)
                .get_variant(v.parent.parent.name)
                .get_branch(v.parent.name)
                .get_version(v.name)
            )
            if self.access_granted(version.url()):
                versions.append(version)
                if len(versions) == num:
                    break
        return versions

    @aiohttp_jinja2.template("home.html")
    async def get(self):
        repo = self.repo()
        user = self.request.match_info.get("user")
        if user:
            if not repo.path().exists():
                raise web.HTTPNotFound()
            if not self.request.path.endswith("/"):
                raise web.HTTPFound(self.request.path + "/")
            return {
                "disk_usage": repo.disk_usage,
                "quota": repo.QUOTA,
                "human_readable": human_readable,
                "latest_releases": await self.get_latest_releases(repo),
                "access": {
                    "branches": self.access_granted(repo.url() + "branches/"),
                    "products": self.access_granted(repo.url() + "products/"),
                },
            }
        users = []
        for r in repo.get_user_repos():
            if self.access_granted(r.url()):
                users.append(r.user)
        return {
            "users": users,
            "latest_releases": await self.get_latest_releases(repo),
            "access": {
                "branches": self.access_granted("/branches/"),
                "products": self.access_granted("/products/"),
            },
        }


# --------------------------------------------------------------------------------------
class StaticView(BaseView):

    ROOT = pathlib.Path(__file__).parent.parent.parent
    PUBLIC_URL = os.getenv("DLREPO_PUBLIC_URL")
    CLI = b""
    for root in (ROOT, pathlib.Path("/usr/local/bin"), pathlib.Path("/usr/bin")):
        cli = root / "dlrepo-cli"
        if cli.is_file():
            CLI = cli.read_text(encoding="utf-8")
            if PUBLIC_URL:
                CLI = CLI.replace("http://127.0.0.1:1337", PUBLIC_URL)
            CLI = CLI.encode("utf-8")
            break
    CLI_HEADERS = {
        "Content-Type": "text/plain; charset=utf-8",
        "Content-Length": str(len(CLI)),
    }
    STATIC_DIRS = []
    if os.getenv("DLREPO_STATIC_DIR"):
        STATIC_DIRS.append(pathlib.Path(os.getenv("DLREPO_STATIC_DIR")))
    STATIC_DIRS.append(ROOT / "dlrepo/static")

    @classmethod
    def urls(cls):
        yield "/cli"
        yield "/static/{file}"

    def resolve_filepath(self):
        relpath = self.request.match_info["file"]
        if relpath.startswith("/") or any(x in (".", "..") for x in relpath.split("/")):
            raise web.HTTPNotFound()
        for static_dir in self.STATIC_DIRS:
            path = static_dir / relpath
            if path.is_file():
                return path
        raise web.HTTPNotFound()

    async def get(self):
        if self.request.path == "/cli":
            return web.Response(body=self.CLI, headers=self.CLI_HEADERS)
        # Do not use self.file_response to avoid X-Sendfile.
        # The static roots are dynamic and it makes complex reverse proxy configs.
        # The static files are small anyway.
        return web.FileResponse(self.resolve_filepath())

    async def head(self):
        if self.request.path == "/cli":
            return web.Response(headers=self.CLI_HEADERS)
        return await self.get()


# --------------------------------------------------------------------------------------
async def template_vars(request):
    return {
        "year": datetime.now().strftime("%Y"),
        "request": request,
        **request.match_info,
    }


# --------------------------------------------------------------------------------------
def pretty_time(timestamp: int, fmt: str = "%Y %b %d, %H:%M:%S UTC") -> str:
    if not timestamp:
        return "n/a"
    utc_time = datetime.fromtimestamp(timestamp, timezone.utc)
    return utc_time.strftime(fmt)


# --------------------------------------------------------------------------------------
def add_routes(app):
    template_dirs = []
    if os.getenv("DLREPO_TEMPLATES_DIR"):
        template_dirs.append(os.getenv("DLREPO_TEMPLATES_DIR"))
    base_dir = os.path.dirname(os.path.dirname(__file__))
    template_dirs += [
        os.path.join(base_dir, "templates"),
        # add the parent path of the default templates dir to allow overriding
        # builtin templates in DLREPO_TEMPLATES_DIR
        # Inspired from https://github.com/ipython/ipython/commit/905835ea53d3a
        base_dir,
    ]
    aiohttp_jinja2.setup(
        app,
        loader=jinja2.FileSystemLoader(template_dirs),
        context_processors=[template_vars],
        extensions=["jinja2.ext.do"],
        trim_blocks=True,
        lstrip_blocks=True,
        filters={"pretty_time": pretty_time},
    )
    for route in (
        HomeView,
        StaticView,
        # artifacts
        BranchesView,
        BranchView,
        TagView,
        JobArchiveView,
        JobView,
        FormatArchiveView,
        FormatDigestsView,
        FormatFileView,
        FormatDirView,
        ArtifactView,
        ProductsView,
        ProductView,
        ProductVariantView,
        ProductBranchView,
        VersionArchiveView,
        VersionView,
        # docker
        RootView,
        CatalogView,
        ManifestView,
        TagsListView,
        NewBlobUploadView,
        BlobsUploadsView,
        BlobsView,
        TagsListView,
        ManifestReadOnlyView,
    ):
        for url in route.urls():
            app.add_routes([web.view(url, route)])
